# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict, Tuple, Any, SupportsFloat


class BinaryMinHeap:
    """Can be used as a priority queue.
    Allows changing priorities."""

    def __init__(self):
        self.heap: List[Tuple[SupportsFloat, Any]] = []
        """The actual heap structure."""
        self.mapping: Dict[Any, int] = dict()
        """Maps objects to positions in the heap."""

    def push(self, val, priority):
        idx = len(self.heap)
        self.heap.append((priority, val))
        self.mapping[val] = idx
        self._sift_up(idx)

    def pop(self):
        head = self.heap[0][1]
        self.mapping.pop(head)
        tmp_head = self.heap.pop()
        if len(self.heap) > 0:
            self.heap[0] = tmp_head
            self.mapping[tmp_head[1]] = 0
            self._sift_down(0)
        return head

    def change_priority(self, val, priority):
        idx = self.mapping[val]
        old_priority = self.heap[idx][0]
        if old_priority == priority:
            return
        self.heap[idx] = (priority, val)
        if priority < old_priority:
            self._sift_up(idx)
        else:
            self._sift_down(idx)

    def _sift_up(self, idx):
        ptr_idx = idx
        ptr = self.heap[ptr_idx]
        while ptr_idx > 0:
            parent_idx = _parent_index(ptr_idx)
            parent = self.heap[parent_idx]
            if ptr[0] < parent[0]:
                self.heap[parent_idx] = ptr
                self.mapping[ptr[1]] = parent_idx
                self.heap[ptr_idx] = parent
                self.mapping[parent[1]] = ptr_idx
                ptr_idx = parent_idx
            else:
                return

    def _sift_down(self, idx):
        ptr_idx = idx
        ptr = self.heap[ptr_idx]
        while True:
            child_idx = _lchild_index(ptr_idx)
            if child_idx >= len(self.heap):
                return
            child = self.heap[child_idx]
            rchild_idx = _rchild_index(ptr_idx)
            if rchild_idx < len(self.heap):
                rchild = self.heap[rchild_idx]
                if rchild[0] < child[0]:
                    child_idx = rchild_idx
                    child = rchild
            if ptr[0] <= child[0]:
                return
            self.heap[child_idx] = ptr
            self.heap[ptr_idx] = child
            self.mapping[child[1]] = ptr_idx
            self.mapping[ptr[1]] = child_idx
            ptr_idx = child_idx


def _parent_index(idx: int):
    return (idx-1) // 2


def _lchild_index(idx: int):
    return 2*idx+1


def _rchild_index(idx: int):
    return 2*idx+2



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Set, Iterable, Tuple, DefaultDict
import rdflib
from scipy.sparse import csr_matrix, vstack
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
from stwfsapy import thesaurus as t
from stwfsapy.util.set_closure import set_closure
from collections import defaultdict


class ThesaurusFeatureTransformation(BaseEstimator, TransformerMixin):
    def __init__(
            self,
            graph: rdflib.Graph,
            concepts: Set[rdflib.term.URIRef],
            thesauri: Set[rdflib.term.URIRef],
            thesaurus_relation: rdflib.term.URIRef,
            inverse_relation: bool = False
            ):
        self.graph = graph
        self.concepts = concepts
        self.thesauri = thesauri
        self.mapping_ = None
        self.thesaurus_relation = thesaurus_relation
        self.inverse_relation = inverse_relation

    def fit(self, X=None, y=None, **kwargs):
        """Creates the mapping from concepts
        to the thesauri that are broader than the concept."""
        broaders = list(t.extract_relation_by_uri(
            self.graph,
            self.thesaurus_relation,
            self.inverse_relation))
        concept_po = _collect_po_from_tuples(
            t.filter_subject_tuples_from_set(broaders, self.concepts))
        thesauri_po = _collect_po_from_tuples(
            t.filter_subject_tuples_from_set(broaders, self.thesauri),
            self.thesauri)
        for thesaurus in self.thesauri:
            thesauri_po[thesaurus].add(thesaurus)
        thesauri_closure = set_closure(thesauri_po)
        thesaurus_indices = dict(zip(
            thesauri_closure,
            range(len(thesauri_closure))))
        concept_thesauri_mapping = {
            concept: set.union(
                *(
                    thesauri_closure.get(broader, set())
                    for broader
                    in broaders
                ))
            for concept, broaders in concept_po.items()
        }
        self.feature_dim_ = max(len(thesaurus_indices), 1)
        self.mapping_ = {
            str(concept): csr_matrix(
                (
                    [1 for _ in thesaurii],
                    (
                        [0 for _ in thesaurii],
                        [
                            thesaurus_indices[thesaurus]
                            for thesaurus
                            in thesaurii]
                    )
                ),
                shape=(1, len(thesaurus_indices))
            )
            for concept, thesaurii
            in concept_thesauri_mapping.items()
        }
        return self

    def _transform_single(self, x):
        # No default dict, so the transform can be pickled
        try:
            res = self.mapping_[x]
        except KeyError:
            res = csr_matrix(
                ([], ([], [])),
                shape=(1, self.feature_dim_)
            )
        return res

    def transform(self, X) -> csr_matrix:
        if self.mapping_ is None:
            raise NotFittedError
        return vstack([self._transform_single(x) for x in X])


def _collect_po_from_tuples(
        tuples: Iterable[Tuple[rdflib.term.URIRef, rdflib.term.URIRef]],
        base_elements: Set[rdflib.term.URIRef] = set()
        ) -> DefaultDict[rdflib.term.URIRef, Set[rdflib.term.URIRef]]:
    ret = defaultdict(set)
    for e in base_elements:
        ret[e].add(e)
    for narrower, broader in tuples:
        ret[narrower].add(broader)
    return ret



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from typing import Tuple, Set, Dict, FrozenSet, Iterable, List
from stwfsapy.automata import dfa
from stwfsapy.automata import nfa
from queue import Queue


class NfaToDfaConverter:
    """Converts a nondeterminisitic finite automaton (NFA)
    into a deterministic one.
    The NFA must be free of empty transitions."""

    def __init__(self, nfa_automaton: nfa.Nfa):
        self.nfa: nfa.Nfa = nfa_automaton
        """The input automaton."""
        self.dfa: dfa.Dfa = dfa.Dfa()
        """The resulting automaton."""
        idx0 = self.dfa.add_state()
        self.queue: Queue = Queue()
        """Queue for controlling the processing order."""
        self.queue.put(idx0)
        start_states = frozenset(self.nfa.starts)
        self.state_represents: Dict[int, List[int]] = {
            idx0: list(start_states)}
        """Maps a state index of the DFA to a set of NFA state indices."""
        self.state_cache: Dict[FrozenSet[int], int] = {start_states: 0}
        """Maps a set of NFA state indices to a DFA state index."""

    def start_conversion(self) -> dfa.Dfa:
        while self.queue.qsize() > 0:
            self.perform_step(self.queue.get())
            self.queue.task_done()
        return self.dfa

    def perform_step(self, dfa_start_state_idx: int):
        (
            symbol_transitions,
            non_word_char_transitions,
            accepts
        ) = self._collect_nfa_transitions(
            self.state_represents[dfa_start_state_idx])
        self._create_dfa_transitions(
            dfa_start_state_idx,
            symbol_transitions,
            non_word_char_transitions,
            accepts)

    def _collect_nfa_transitions(self, states: Iterable[int]) -> Tuple:
        symbol_transitions = defaultdict(set)
        non_word_char_transitions = set()
        accepts = set()
        for nfa_start_idx in states:
            for (
                symbol,
                nfa_end_state_idxs) in (
                    self.nfa.states[nfa_start_idx].symbol_transitions.items()):
                symbol_transitions[symbol].update(nfa_end_state_idxs)
            non_word_char_transitions.update(
                    self.nfa.states[nfa_start_idx].non_word_char_transitions)
            accepts.update(self.nfa.states[nfa_start_idx].accepts)
        return symbol_transitions, non_word_char_transitions, accepts

    def _create_dfa_transitions(
            self,
            dfa_start_state_idx: int,
            symbol_transitions: Dict[str, Set[int]],
            non_word_char_transitions: Set[int],
            accepts: Set[object]):
        for symbol, nfa_end_state_idxs in symbol_transitions.items():
            dfa_end_state_idx = self._get_or_create_dfa_state(
                nfa_end_state_idxs)
            self.dfa.set_symbol_transition(
                dfa_start_state_idx,
                dfa_end_state_idx,
                symbol)
        if len(non_word_char_transitions) > 0:
            dfa_end_state_idx = self._get_or_create_dfa_state(
                non_word_char_transitions)
            self.dfa.set_non_word_char_transition(
                dfa_start_state_idx,
                dfa_end_state_idx)
        if len(accepts) > 0:
            self.dfa.add_acceptances(dfa_start_state_idx, accepts)

    def _get_or_create_dfa_state(self, state_idxs: Iterable[int]) -> int:
        """Retrieves a state index of the DFA representing a set of
        state indices in the NFA. If such a state does not exist,
        a new one will be created. The new State will also be added to
        the state lookup tables."""
        frozen_states = frozenset(state_idxs)
        try:
            dfa_state_idx = self.state_cache[frozen_states]
        except KeyError:
            dfa_state_idx = self.dfa.add_state()
            self.queue.put(dfa_state_idx)
            self.state_represents[dfa_state_idx] = list(frozen_states)
            self.state_cache[frozen_states] = dfa_state_idx
            acceptance: Set = set()
            for state_idx in state_idxs:
                acceptance.union(self.nfa.states[state_idx].accepts)
            dfa_state = self.dfa.states[dfa_state_idx]
            dfa_state.accepts = list(acceptance)

        return dfa_state_idx



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict, Iterable, Tuple, Any, Callable, Optional

_KEY_STATE_SYMBOL_TRANSITIONS = 'symbol_transitions'
_KEY_STATE_NON_WORD_CHAR_TRANSITION = 'non_word_char_transitions'
_KEY_STATE_ACCEPTS = 'accepts'
_KEY_DFA_STATES = 'states'


class State:

    def __init__(
            self,
            symbol_transitions=None,
            non_word_char_transition=None,
            accepts=None):
        self.symbol_transitions: Dict[str, int] = symbol_transitions or dict()
        """Transitions consuming a symbol."""
        self.non_word_char_transition: Optional[int] = non_word_char_transition
        """Transition consuming a non word character symbol."""
        self.accepts: List[Any] = accepts or []
        """What is accepted by this state."""

    def set_symbol_transition(self, symbol: str, idx: int):
        self.symbol_transitions[symbol] = idx

    def set_non_word_char_transition(self, idx: int):
        self.non_word_char_transition = idx

    def add_acceptances(self, accepts: Iterable):
        self.accepts.extend(accepts)

    def to_dict(
            self,
            acceptance_handler: Callable) -> Dict[str, Any]:
        return {
            _KEY_STATE_SYMBOL_TRANSITIONS: self.symbol_transitions,
            _KEY_STATE_NON_WORD_CHAR_TRANSITION: self.non_word_char_transition,
            _KEY_STATE_ACCEPTS: [acceptance_handler(a) for a in self.accepts],
        }

    def __eq__(self, other):
        return isinstance(other, State) and self.accepts == other.accepts and (
            self.symbol_transitions == other.symbol_transitions) and (
                self.non_word_char_transition == other.non_word_char_transition
            )

    @staticmethod
    def from_dict(
            conf: Dict[str, Any],
            acceptance_handler: Callable):
        return State(
            symbol_transitions=conf[_KEY_STATE_SYMBOL_TRANSITIONS],
            non_word_char_transition=conf.get(
                _KEY_STATE_NON_WORD_CHAR_TRANSITION),
            accepts=[
                acceptance_handler(accept)
                for accept
                in conf[_KEY_STATE_ACCEPTS]
            ]
        )


class Dfa:
    """Represents a deterministic finite automaton.
    This class is not intended to be used directly.
    Instead create a stwfsapy.automata.nfa.Nfa and convert it.
    Example:
        from stwfsapy.automata import nfa
        from stwfsapy.automata import conversion

        nfautomaton = nfa.Nfa()

        # Construct automaton,
        # e.g., using stwfsapy.automata.construction
        # ...

        nfa.remove_empty_transitions()

        dfa = conversion.NfaToDfaConverter(nfautomaton).start_conversion()"""

    def __init__(self, states=None):
        self.states: List[State] = states or []
        """All the states of this Automaton."""

    def add_state(self) -> int:
        """Creates a new state and returns its index."""
        idx = len(self.states)
        state = State()
        self.states.append(state)
        return idx

    def add_acceptances(self, idx: int, accepts: Iterable):
        """Adds new acceptances to a state.
        The state is given by its index."""
        self.states[idx].add_acceptances(accepts)

    def set_symbol_transition(self, start_idx: int, end_idx: int, symbol: str):
        """Add a new symbol consuming transition between to states.
        The states are given by their index."""
        self.states[start_idx].set_symbol_transition(symbol, end_idx)

    def set_non_word_char_transition(self, start_idx: int, end_idx: int):
        """Add a new non word char consuming transition between two states.
        The states are given by their index."""
        self.states[start_idx].set_non_word_char_transition(end_idx)

    def search(self, text: str) -> Iterable[Tuple[Any, str, int, int]]:
        """ Process a string, yielding acceptances of all substrings.
        The method assumes a DFA which has been constructed by the application
        of stwfsapy.automata.construction.ConstructionState and
        stwfsapy.automata.conversion.NfaToDfaConverter."""
        # At construction time we add non word char transitions at
        # the beginning and end of a label. Therefore add them for search.
        search_text = f'.{text}.'
        last_end_position = 0
        for start in range(len(search_text)):
            if start < last_end_position:
                continue
            stack = [(0, start)]
            while len(stack) > 0:
                state_idx, position = stack.pop()
                state = self.states[state_idx]
                non_empty_accepts = False
                for accept in state.accepts:
                    last_end_position = position
                    # subtract one from position,
                    # as the match is without consuming the next symbol
                    # subtract another one for the '.' introduced at the start
                    original_end = position - 2
                    yield (
                        accept,
                        text[start:original_end],
                        start,
                        original_end)
                    non_empty_accepts = True
                if non_empty_accepts:
                    break
                try:
                    symbol = search_text[position]
                except IndexError:
                    continue
                # First append non word_char.
                # As stack is LIFO,
                # explicit symbol transitions will be preferred.
                if not symbol.isalnum() and state.non_word_char_transition:
                    stack.append(
                        (state.non_word_char_transition, position+1))
                try:
                    transition = state.symbol_transitions[symbol]
                    stack.append((transition, position+1))
                except KeyError:
                    pass

    def to_dict(
            self,
            acceptance_handler: Callable) -> Dict[str, Any]:
        return {
            _KEY_DFA_STATES: [
                state.to_dict(acceptance_handler)
                for state
                in self.states
            ]
        }

    @staticmethod
    def from_dict(
            conf: Dict[str, Any],
            acceptance_handler: Callable):
        return Dfa([
            State.from_dict(state_conf, acceptance_handler)
            for state_conf
            in conf[_KEY_DFA_STATES]
        ])

    def __eq__(self, other):
        return isinstance(other, Dfa) and self.states == other.states



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Set, Dict, Hashable, Tuple, List


def set_closure(
        sets: Dict[Hashable, Set[Hashable]]
        ) -> Dict[Hashable, Set[Hashable]]:
    """Computes the closure for each element of a antisymmetric relation.
    The relation is given by a mapping from elements
    to a set of related elements.
    Raises a RelationLoopException if the relation has circles."""
    closure: Dict[Hashable, Set[Hashable]] = dict()
    for key in sets:
        stack: List[Tuple[Hashable, Set[Hashable]]] = [(key, set())]
        while len(stack) > 0:
            top = stack[-1]
            current = top[0]
            ancestors = top[1]
            if current in closure:
                stack.pop()
            else:
                if current not in sets:
                    stack.pop()
                    closure[current] = set()
                else:
                    all_known = True
                    for child in sets[current]:
                        if child not in closure and child != current:
                            if child in ancestors:
                                raise RelationLoopException()
                            all_known = False
                            new_ancestors = ancestors.copy()
                            new_ancestors.add(current)
                            stack.append((child, new_ancestors))
                    if all_known:
                        element_closure = set()
                        for child in sets[current]:
                            element_closure.add(child)
                            try:
                                for closure_object in closure[child]:
                                    element_closure.add(closure_object)
                            except KeyError:
                                pass
                        closure[current] = element_closure
                        stack.pop()
    return closure


class RelationLoopException(Exception):
    pass



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from typing import Set, List, Any, DefaultDict
from stwfsapy.automata.heap import BinaryMinHeap


class State:

    def __init__(self):
        self.symbol_transitions: DefaultDict[str, Set[int]] = defaultdict(set)
        """Determines the states that can be reached
        by consuming a character."""
        self.non_word_char_transitions: Set[int] = set()
        """Is not None if there is a transition via a non word symbol."""
        self.empty_transitions: Set[int] = set()
        """What can be reached using the empty String/Symbol."""
        self.incoming_empty_transitions: Set[int] = set()
        """How this state can be reached using the empty String/Symbol."""
        self.accepts: List[Any] = []
        """What is accepted by this State."""


class Nfa:

    def __init__(self):
        self.states: List[State] = []
        """All states of the NFA's graph"""
        self.starts: List[int] = []
        """All start states."""

    def add_state(self) -> int:
        """Creates a new state and returns its index."""
        idx = len(self.states)
        self.states.append(State())
        return idx

    def add_start(self, idx: int):
        """Adds a state, by index to the list of possible start states."""
        self.starts.append(idx)

    def add_acceptance(self, idx, accept):
        """Add acceptance to a state. The state is identified by its index."""
        self.states[idx].accepts.append(accept)

    def add_symbol_transition(self, start: int, end: int, symbol: str):
        """Add a symbol consuming transition between two states.
        The states are represented by their indices."""
        transitions = self.states[start].symbol_transitions
        transitions[symbol].add(end)

    def add_empty_transition(self, start: int, end: int):
        """Add an empty Transition between two states.
        This transition does not consume a symbol.
        The states are represented by their indices."""
        self.states[start].empty_transitions.add(end)
        self.states[end].incoming_empty_transitions.add(start)

    def add_non_word_char_transition(self, start: int, end: int):
        """Add a non word char consuming transition between two states.
        The states are represented by their indices."""
        self.states[start].non_word_char_transitions.add(end)

    def remove_empty_transitions(self):
        """Removes all empty transitions in the NFA.
        Does not support loops of empty transitions."""
        queue = BinaryMinHeap()
        for idx in range(len(self.states)):
            if len(self.states[idx].incoming_empty_transitions) > 0:
                queue.push(idx, len(self.states[idx].empty_transitions))
        while len(queue.heap) > 0:
            ptr_idx = queue.pop()
            ptr = self.states[ptr_idx]
            if len(ptr.empty_transitions) > 0:
                raise Exception(
                    "There is an empty transition loop in the NFA.")
            for incoming_idx in ptr.incoming_empty_transitions.copy():
                incoming = self.states[incoming_idx]
                self._remove_empty_transition(
                    incoming_idx,
                    incoming,
                    ptr_idx,
                    ptr)
                if len(incoming.incoming_empty_transitions) > 0:
                    queue.change_priority(
                        incoming_idx,
                        len(incoming.empty_transitions))

    def _remove_empty_transition(
            self,
            start_idx: int,
            start: State,
            end_idx: int,
            end: State):
        for symbol, states in end.symbol_transitions.items():
            for state_idx in states:
                start.symbol_transitions[symbol].add(state_idx)
        for state_idx in end.non_word_char_transitions:
            start.non_word_char_transitions.add(state_idx)
        start.empty_transitions.discard(end_idx)
        end.incoming_empty_transitions.discard(start_idx)



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List
from stwfsapy.automata import nfa


class ConstructionState:

    def __init__(self, graph: nfa.Nfa, expression: str, accept: object):
        self.graph: nfa.Nfa = graph
        """The graph the expression will be added to."""
        self.expression: str = expression
        """The expression that will be added to the graph."""
        self.accept: object = accept
        """Object returned when the expression matches an input."""

    def _set_up(self):
        self.start_idx = self.graph.add_state()
        self.expression_start_idx: int = self.graph.add_state()
        """Start node of the expression."""
        self.graph.add_non_word_char_transition(
            self.start_idx,
            self.expression_start_idx)
        alternation_start_idx = self.graph.add_state()
        self.graph.add_empty_transition(
            self.expression_start_idx,
            alternation_start_idx)
        self.append_to: List[int] = [alternation_start_idx]
        """States that are the start of the next transition."""
        self.before_braces: List[List[int]] = [
            [self.expression_start_idx],
            [alternation_start_idx]]
        """Stack of pointers to states directly preceding an opening brace."""
        self.dangling_alternations: _AlternationManager = _AlternationManager()
        """Handles the end of alternations."""
        self.escape_next: bool = False
        """Indicates whether the next symbol should be taken literally.
        Even if it is a control symbol."""

    def construct(self):
        """Starts the construction process."""
        self._set_up()
        for i in range(len(self.expression)):
            self._perform_step(i)
        ends = self.append_to
        try:
            danglings = self.dangling_alternations.pop()
            if danglings != [self.expression_start_idx]:
                ends.extend(danglings)
        except IndexError:
            pass
        for end_idx in ends:
            acceptance_idx = self.graph.add_state()
            self.graph.add_acceptance(acceptance_idx, self.accept)
            self.graph.add_non_word_char_transition(end_idx, acceptance_idx)
        self.graph.add_start(self.start_idx)

    def _perform_step(self, idx: int):
        """Consumes a single input character of the input expression."""
        symbol = self.expression[idx]
        if self.escape_next:
            self.escape_next = False
            self._process_symbol(symbol)
        elif symbol == "(":
            self._process_opening_brace()
        elif symbol == ")":
            self._process_closing_brace()
        elif symbol == "?":
            self._process_optional()
        elif symbol == "*":
            self._process_kleene_closure()
        elif symbol == "\\":
            self.escape_next = True
        elif symbol == "|":
            self._process_alternation()
        else:
            self._process_symbol(symbol)

    def _process_symbol(self, symbol):
        new_state_idx = self.graph.add_state()
        for state_idx in self.append_to:
            self.graph.add_symbol_transition(state_idx, new_state_idx, symbol)
        self.before_braces[-1] = self.append_to
        self.append_to = [new_state_idx]

    def _process_opening_brace(self):
        new_state_idx = self.graph.add_state()
        for state_idx in self.append_to:
            self.graph.add_empty_transition(state_idx, new_state_idx)
        self.dangling_alternations.push_empty()
        self.before_braces[-1] = self.append_to
        self.before_braces.append([new_state_idx])
        self.append_to = [new_state_idx]

    def _process_alternation(self):
        self.dangling_alternations.push(self.append_to)
        before_braces = self.before_braces[-2]
        new_state = self.graph.add_state()
        for before_brace in before_braces:
            self.graph.add_empty_transition(before_brace, new_state)
        self.before_braces[-1] = [new_state]
        self.append_to = [new_state]

    def _process_closing_brace(self):
        danglings = self.dangling_alternations.pop()
        self.append_to.extend(danglings)
        self.before_braces.pop()

    def _process_optional(self):
        before_idxs = self.before_braces[-1]
        for before_idx in before_idxs:
            for state_idx in self.append_to:
                self.graph.add_empty_transition(before_idx, state_idx)

    def _process_kleene_closure(self):
        before_idxs = self.before_braces[-1]
        after_idx = self.graph.add_state()
        for state_idx in self.append_to:
            self.graph.add_empty_transition(state_idx, after_idx)
        for before_idx in before_idxs:
            self.graph.add_empty_transition(before_idx, after_idx)
            for state_idx in self.append_to:
                self.graph.add_empty_transition(state_idx, before_idx)
        self.append_to = [after_idx]


class _AlternationManager():

    def __init__(self):
        self.stack = []

    def pop(self) -> List[int]:
        return self.stack.pop()

    def push(self, states: List[int]):
        try:
            self.stack[-1].extend(states)
        except IndexError:
            self.stack.append(states)

    def push_empty(self):
        self.stack.append([])



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.pipeline import FeatureUnion
from sklearn.exceptions import NotFittedError
import numpy as np
import re


_NAME_VECTOR_FEATURE = "vectorizer"
_NAME_CHAR_FEATURE = "n_chars"
_NAME_WORD_FEATURE = "n_words"
_NAME_SPECIAL_CHARS_FEATURE = "n_special"
_NAME_UPPER_FEATURE = "n_upper"
_NAME_DIGIT_FEATURE = "n_digits"


def mk_text_features():
    return FeatureUnion([
        (_NAME_VECTOR_FEATURE, CountVectorizer(lowercase=False)),
        (_NAME_CHAR_FEATURE, CountFeature(CountType.N_CHAR)),
        (_NAME_WORD_FEATURE, CountFeature(CountType.N_WORD)),
        (_NAME_SPECIAL_CHARS_FEATURE, CountFeature(CountType.N_SPECIAL)),
        (_NAME_UPPER_FEATURE, CountFeature(CountType.N_UPPER)),
        (_NAME_DIGIT_FEATURE, CountFeature(CountType.N_DIGIT)),
    ])


class CountType(Enum):
    N_CHAR = 0
    N_WORD = 1
    N_SPECIAL = 2
    N_UPPER = 3
    N_DIGIT = 4


class CountFeature(BaseEstimator, TransformerMixin):
    def __init__(self, ctype: CountType):
        self.ctype = ctype
        self.function_ = None

    def fit(self, X, y=None):
        self.function_ = self.fun_from_type(self.ctype)
        return self

    def transform(self, X):
        if not self.function_:
            raise NotFittedError
        ret = np.empty((len(X), 1))
        for i, x in enumerate(X):
            ret[i, 0] = self.function_(x)
        return ret

    @staticmethod
    def fun_from_type(ctype: CountType):
        if ctype == CountType.N_CHAR:
            return _count_char
        elif ctype == CountType.N_WORD:
            return _count_word
        elif ctype == CountType.N_SPECIAL:
            return _count_special
        elif ctype == CountType.N_UPPER:
            return _count_upper
        elif ctype == CountType.N_DIGIT:
            return _count_digit


_re_special = re.compile(r"""["'?!()]""")
_re_digit = re.compile(r"\d")


def _count_char(txt: str):
    return len(txt)


def _count_word(txt: str):
    return txt.count(" ")


def _count_special(txt: str):
    return len(_re_special.findall(txt))


def _count_upper(txt: str):
    return sum((c.isupper() for c in txt))


def _count_digit(txt: str):
    return len(_re_digit.findall(txt))



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
file_dir = pathlib.Path(__file__).parent

# The text of the README file
readme_txt = (file_dir / "Readme.md").read_text()

setup(
    name="stwfsapy",
    version="0.01.4",
    description="A library for match labels of thesaurus concepts to text" + (
        " and assigning scores to found occurrences."),
    long_description=readme_txt,
    long_description_content_type="text/markdown",
    url="https://github.com/zbw/stwfsapy",
    author="Moritz Fuerneisen",
    author_email="m.fuerneisen@zbw.eu",
    license="Apache",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["rdflib>=4.2,<6.0", "scikit-learn==0.22.*"],
    tests_require=['py', 'pytest', 'pytest-mock'],
    extras_require={
        'dev': [
            "pytest==5.4.*",
            "pytest-mock==1.7.*",
            "pytest-pycodestyle==2.2.*",
            "pytest-cov==2.8.*",
            "codecov==2.1.*",
            ]
        },
    python_requires='>=3.6',
)



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Tuple, FrozenSet, Iterable, Any, Optional
from rdflib import Graph
from rdflib.term import Literal, URIRef
from rdflib.namespace import SKOS, OWL, RDF


def extract_labels(g: Graph) -> Iterable[Tuple[URIRef, Literal]]:
    """
    Extracts SKOS.prefLabels and SKOS.altLabels from a rdflib.Graph
    """
    return g[: SKOS.prefLabel | SKOS.altLabel]


def extract_by_type_uri(
        g: Graph,
        type_URI: URIRef,
        remove: Optional[FrozenSet[URIRef]] = None) -> Iterable[URIRef]:
    """Extract all elements of a specific type from a rdflib graph.
    Allows to exclude the elements that are in a specified set."""
    by_type = g[:RDF.type:type_URI]
    if not remove:
        return by_type
    else:
        return _filter_refs_from_set_complement(by_type, remove)


def extract_deprecated(g: Graph):
    return g[:OWL.deprecated: Literal(True)]


def extract_relation_by_uri(g: Graph, uri: URIRef, reverse: bool):
    subj_objs = g[:uri:]
    if reverse:
        return map(lambda x: (x[1], x[0]), subj_objs)
    else:
        return subj_objs


def _predicate_uri_from_set(uri: URIRef, uri_set: FrozenSet[URIRef]):
    return uri in uri_set


def filter_subject_tuples_from_set(
        tuples: Iterable[Tuple[URIRef, Any]],
        uri_set: FrozenSet[URIRef]
        ) -> Iterable[Tuple[URIRef, Any]]:
    """Filters an iterable of tuples.
    Keeps only tuples where the first element is
    present in the provided set."""
    return filter(lambda t: _predicate_uri_from_set(t[0], uri_set), tuples)


def _predicate_refs_from_set_complement(
        uri: URIRef,
        uri_set: FrozenSet[URIRef]):
    return uri not in uri_set


def _filter_refs_from_set_complement(
        tuples: Iterable[URIRef],
        uri_set: FrozenSet[URIRef],
        ):
    return filter(
        lambda t: _predicate_refs_from_set_complement(t, uri_set),
        tuples)


def _predicate_langs(lit: Literal, langs: FrozenSet[str]) -> bool:
    return lit.language in langs


def _filter_by_langs(
        tuples: Iterable[Tuple[URIRef, Literal]],
        langs: FrozenSet[str]
        ) -> Iterable[Tuple[URIRef, Literal]]:
    return filter(lambda t: _predicate_langs(t[1], langs), tuples)


def _unwrap_label(tuple: Tuple[URIRef, Literal]) -> Tuple[URIRef, str]:
    return (tuple[0], tuple[1].value)


def _unwrap_labels(
        tuples: Iterable[Tuple[URIRef, Literal]]
        ) -> Iterable[Tuple[URIRef, str]]:
    return map(_unwrap_label, tuples)


def retrieve_concept_labels(
        g: Graph,
        allowed: Optional[FrozenSet[URIRef]] = frozenset(),
        langs: FrozenSet[str] = frozenset()
        ) -> Iterable[Tuple[URIRef, str]]:
    """Extracts altLabels and prefLabels from a SKOS graph.

    Only the labels that are in one of the specified language will be reported.
    In addition the concept URIs can be limited by a set.
    Args:
        g: The SKOS graph whose labels are extracted.

        allowed: Only concepts present in the set are retained.
            If None or the set is empty
            all concepts will be present in the result.


        langs: Only retain labels that are in the given language.
            Add None to the set if you want to keep labels
            without a language annotation.


    Returns:
        An iterator of pairs. The first element is a URIRef,
        representing a concept.
        The second element is a label for the concept.

        """
    refs_with_labels = extract_labels(g)
    if langs is not None and len(langs) > 0:
        filtered_by_language = _filter_by_langs(refs_with_labels, langs)
    else:
        filtered_by_language = refs_with_labels
    if allowed:
        filtered_by_set = filter_subject_tuples_from_set(
            filtered_by_language,
            allowed)
    else:
        filtered_by_set = filtered_by_language
    unwrapped_labels = _unwrap_labels(filtered_by_set)
    return unwrapped_labels



# Copyright 2020 Leibniz Information Centre for Economics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import FrozenSet, List, Iterable, Container, Tuple, TypeVar, Union
from logging import getLogger
from rdflib.term import URIRef
from rdflib import Graph
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.tree import DecisionTreeClassifier
from scipy.sparse import csr_matrix
from stwfsapy import thesaurus as t
from stwfsapy.automata import nfa, construction, conversion, dfa
from stwfsapy.thesaurus_features import ThesaurusFeatureTransformation
from stwfsapy.text_features import mk_text_features
from stwfsapy import case_handlers
from stwfsapy import expansion
import pickle as pkl
from json import dumps, loads
from zipfile import ZipFile


T = TypeVar('T')
N = TypeVar('N', int, float)
Nl = TypeVar('Nl', List[int], List[float])


_KEY_DFA = 'dfa'
_KEY_PIPELINE = 'pipeline'
_KEY_CONCEPT_MAP = 'concept_map'
_KEY_GRAPH_ITEMS = 'graph_items'
_KEY_CONCEPT_TYPE_URI = 'concept_type_uri'
_KEY_THESAURUS_TYPE_URI = 'thesaurus_type_uri'
_KEY_THESAURUS_RELATION_TYPE_URI = 'thesaurus_relation_type_uri'
_KEY_THESAURUS_RELATION_IS_SPECIALISATION = (
    'thesaurus_relation_is_specialisation')
_KEY_REMOVE_DEPRECATED = 'remove_deprecated'
_KEY_LANGS = 'langs'
_KEY_HANDLE_TITLE_CASE = 'handle_title_case'
_KEY_EXTRACT_UPPER_CASE_FROM_BRACES = 'extract_upper_case_from_braces'
_KEY_EXTRACT_ANY_CASE_FROM_BRACES = 'extract_any_key_from_braces'
_KEY_EXPAND_AMPERSAND_WITH_SPACES = 'expand_ampersand_with_spaces'
_KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION = (
    'expand_abbreviation_with_punctuation')
_KEY_SIMPLE_ENGLISH_PLURAL_RULES = 'simple_english_plural_rules'

_NAME_GRAPH_FILE = 'graph.rdf'
_NAME_PIPELINE_FILE = 'pipeline.pkl'
_NAME_PREDICTOR_FILE = 'predictor.json'

_logger = getLogger('stwfsa')


class StwfsapyPredictor(BaseEstimator, ClassifierMixin):
    """Finds labels of thesaurus concepts in texts
    and assigns them a score."""
    def __init__(
            self,
            graph: Graph,
            concept_type_uri: Union[str, URIRef],
            sub_thesaurus_type_uri: Union[str, URIRef] = "",
            thesaurus_relation_type_uri: Union[str, URIRef] = "",
            thesaurus_relation_is_specialisation: bool = False,
            remove_deprecated: bool = True,
            langs: FrozenSet[str] = frozenset(),
            handle_title_case: bool = True,
            extract_upper_case_from_braces: bool = True,
            extract_any_case_from_braces: bool = False,
            expand_ampersand_with_spaces: bool = True,
            expand_abbreviation_with_punctuation: bool = True,
            simple_english_plural_rules: bool = False,
            ):
        """Creates the predictor.
        Args:
            graph: The SKOS onthology used to extract the labels.
            concept_type_uri
                The uri of the concept type.
                It is assumed that for every concept c,
                there is a triple (c, RDF.type, concept_type_uri)
                in the graph.
            sub_thesaurus_type_uri: The uri of the concept type.
                It is assumed that for every sub thesaurus t,
                there is a triple (t, RDF.type, sub_thesaurus_type_uri)
                in the graph.
            thesaurus_relation_type_uri:
                Uri of the relation that links concepts to thesauri.
            thesaurus_relation_is_specialisation:
                Indicates whether the thesaurus_relation links thesauri to
                concepts or the other way round.
                E.g., for the relation skos:broader it should be false.
                Conversely it should be true for skos:narrower.
            remove_deprecated: When True will discard deprecated subjects.
                Deprecation of a subject has to be indicated by
                a triple (s, OWL.deprecated, Literal(True)) in the graph.
            langs: For each language present in the set,
                labels will be extracted from the graph.
                An empy set or None will extract labels regardless of language.
            handle_title_case: When True, will also match labels in title case.
                I.e., in a text the first letter of every word can be upper
                or lower case and will still be matched.
                When False only the case of the first word's first letter
                will be adapted.
                Example:
                    * Given a label "garbage can" and the
                        title "Oscar Lives in a Garbage Can"
                    * When handle_title_case == True
                        the label will match the text.
                    * When handle_title_case == False
                        the label will not match the text.
                        It would however still match
                        "Garbage can is home to grouchy neighbor."."""
        self.graph = graph
        if isinstance(concept_type_uri, str):
            concept_type_uri = URIRef(concept_type_uri)
        self.concept_type_uri = concept_type_uri
        if isinstance(sub_thesaurus_type_uri, str):
            sub_thesaurus_type_uri = URIRef(sub_thesaurus_type_uri)
        self.sub_thesaurus_type_uri = sub_thesaurus_type_uri
        if isinstance(thesaurus_relation_type_uri, str):
            thesaurus_relation_type_uri = URIRef(thesaurus_relation_type_uri)
        self.thesaurus_relation_type_uri = thesaurus_relation_type_uri
        self.thesaurus_relation_is_specialisation = (
            thesaurus_relation_is_specialisation)
        self.remove_deprecated = remove_deprecated
        self.langs = langs
        self.handle_title_case = handle_title_case
        self.extract_upper_case_from_braces = extract_upper_case_from_braces
        self.extract_any_case_from_braces = extract_any_case_from_braces
        self.expand_ampersand_with_spaces = expand_ampersand_with_spaces
        self.expand_abbreviation_with_punctuation = \
            expand_abbreviation_with_punctuation
        self.simple_english_plural_rules = simple_english_plural_rules

    def _init(self):
        all_deprecated = set(t.extract_deprecated(self.graph))
        concepts = set(t.extract_by_type_uri(
            self.graph,
            self.concept_type_uri,
            remove=all_deprecated))
        thesauri = set(t.extract_by_type_uri(
            self.graph,
            self.sub_thesaurus_type_uri,
            remove=all_deprecated
        ))
        self.concept_map_ = dict(zip(map(str, concepts), range(len(concepts))))
        thesaurus_features = ThesaurusFeatureTransformation(
            self.graph,
            concepts,
            thesauri,
            self.thesaurus_relation_type_uri,
            self.thesaurus_relation_is_specialisation
        )
        labels = t.retrieve_concept_labels(
            self.graph,
            allowed=concepts,
            langs=self.langs)
        nfautomat = nfa.Nfa()
        if self.handle_title_case:
            case_handler = case_handlers.title_case_handler
        else:
            case_handler = case_handlers.sentence_case_handler
        expansion_funs = expansion.collect_expansion_functions(
            extract_upper_case_from_braces=self.extract_upper_case_from_braces,
            extract_any_case_from_braces=self.extract_any_case_from_braces,
            expand_ampersand_with_spaces=self.expand_ampersand_with_spaces,
            expand_abbreviation_with_punctuation=(
                self.expand_abbreviation_with_punctuation),
        )
        if self.simple_english_plural_rules:
            plural_fun = expansion.simple_english_plural_fun
        else:
            def plural_fun(x):
                return x
        for concept, label in labels:
            expanded = label
            for f in expansion_funs:
                expanded = f(expanded)
            _handle_construction(
                construction.ConstructionState(
                    nfautomat,
                    plural_fun(case_handler(expanded)),
                    str(concept)
                ),
                concept,
                label)
        nfautomat.remove_empty_transitions()
        converter = conversion.NfaToDfaConverter(nfautomat)
        self.dfa_ = converter.start_conversion()
        self.pipeline_ = Pipeline([
            ("Combined Features", ColumnTransformer([
                ("Thesaurus Features", thesaurus_features, 0),
                ("Text Features", mk_text_features(), 1)])),
            ("Classifier", DecisionTreeClassifier(
                min_samples_leaf=25,
                max_leaf_nodes=100))
        ])

    def fit(self, X, y=None, **kwargs):
        self._init()
        return self._fit_after_init(X, y=y)

    def _fit_after_init(self, X, y=None):
        matches, train_y = self.match_and_extend(X, y)
        self.pipeline_.fit(matches, y=train_y)
        return self

    def predict_proba(self, X) -> csr_matrix:
        match_X, doc_counts = self.match_and_extend(X)
        if match_X:
            predictions = self.pipeline_.predict_proba(match_X)[:, 1]
        else:
            predictions = []
        return self._create_sparse_matrix(
            predictions,
            match_X,
            doc_counts
        )

    def suggest_proba(
            self,
            texts
            ) -> List[List[Tuple[str, float]]]:
        """For a given list of texts,
        this method returns the matched concepts and their scores."""
        match_X, doc_counts = self.match_and_extend(texts)
        if match_X:
            predictions = self.pipeline_.predict_proba(match_X)[:, 1]
        else:
            predictions = []
        combined = StwfsapyPredictor._collect_prediction_results(
            predictions,
            match_X,
            doc_counts
        )
        return [
            [
                (concept, score)
                for concept, score
                in zip(concepts, scores)
            ]
            for concepts, scores
            in combined
        ]

    def predict(self, X) -> csr_matrix:
        match_X, doc_counts = self.match_and_extend(X)
        if match_X:
            predictions = self.pipeline_.predict(match_X)
        else:
            predictions = []
        return self._create_sparse_matrix(
            predictions,
            match_X,
            doc_counts
        )

    def _create_sparse_matrix(
            self,
            values: Nl,
            tuples: List[Tuple[T, str]],
            doc_counts: List[int]
            ) -> csr_matrix:
        return csr_matrix(
            (
                values,
                (
                    [
                        doc_idx
                        for doc_idx
                        in range(len(doc_counts))
                        for _
                        in range(doc_counts[doc_idx])
                        ],
                    [
                        self.concept_map_.get(tpl[0])
                        for tpl
                        in tuples
                    ]
                )),
            shape=(len(doc_counts), len(self.concept_map_))
        )

    @staticmethod
    def _collect_prediction_results(
            values: Nl,
            tuples: List[Tuple[T, str]],
            doc_counts: List[int]
            ) -> List[Tuple[List[T], Nl]]:
        ret = []
        start = 0
        for count in doc_counts:
            end = start+count
            ret.append(([t[0] for t in tuples[start:end]], values[start:end]))
            start = end
        return ret

    def match_and_extend(
            self,
            texts: Iterable[str],
            truth_refss: Iterable[Container] = None
            ) -> Tuple[List[Tuple[str, str]], List[int]]:
        """Retrieves concepts by their labels from text.
        If ground truth values are present,
        it will also return a list of labels for scoring matches.
        If no ground truth values are present, a list
        with the number of matched concepts for each document is returned."""
        concepts = []
        if truth_refss is not None:
            ret_y = []
            for text, truth_refs in zip(texts, map(str, truth_refss)):
                for match in self.dfa_.search(text):
                    concept = match[0]
                    text = match[1]
                    concepts.append((concept, text))
                    ret_y.append(int(concept in truth_refs))
            return concepts, ret_y
        else:
            doc_counts: List[int] = []
            for text in texts:
                count = 0
                for match in self.dfa_.search(text):
                    count += 1
                    concept = match[0]
                    text = match[1]
                    concepts.append((concept, text))
                doc_counts.append(count)
            return concepts, doc_counts

    def store(self, path):
        with ZipFile(path, 'w') as zfile:
            with zfile.open(_NAME_PREDICTOR_FILE, 'w') as fp:
                fp.write(
                    dumps({
                        _KEY_DFA: self.dfa_.to_dict(str),
                        _KEY_CONCEPT_MAP: self.concept_map_,
                        _KEY_CONCEPT_TYPE_URI: _store_uri_ref(
                            self.concept_type_uri),
                        _KEY_THESAURUS_TYPE_URI: _store_uri_ref(
                            self.sub_thesaurus_type_uri),
                        _KEY_THESAURUS_RELATION_TYPE_URI: _store_uri_ref(
                            self.thesaurus_relation_type_uri),
                        _KEY_THESAURUS_RELATION_IS_SPECIALISATION: (
                                self.thesaurus_relation_is_specialisation),
                        _KEY_REMOVE_DEPRECATED: self.remove_deprecated,
                        _KEY_LANGS: list(self.langs),
                        _KEY_HANDLE_TITLE_CASE: self.handle_title_case,
                        _KEY_EXTRACT_UPPER_CASE_FROM_BRACES: (
                            self.extract_upper_case_from_braces),
                        _KEY_EXTRACT_ANY_CASE_FROM_BRACES: (
                            self.extract_any_case_from_braces),
                        _KEY_EXPAND_AMPERSAND_WITH_SPACES: (
                            self.expand_ampersand_with_spaces),
                        _KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION: (
                            self.expand_abbreviation_with_punctuation),
                        _KEY_SIMPLE_ENGLISH_PLURAL_RULES: (
                            self.simple_english_plural_rules),
                        },
                        ensure_ascii=False
                    ).encode('utf-8')
                )
            with zfile.open(_NAME_PIPELINE_FILE, 'w') as fp:
                # No good way to serialize sk-learn classifier,
                # apart from insecure pickling
                pkl.dump(self.pipeline_, fp)
            with zfile.open(_NAME_GRAPH_FILE, 'w') as fp:
                fp.write(self.graph.serialize(encoding='utf-8'))

    @staticmethod
    def load(path):
        with ZipFile(path, 'r') as zfile:
            with zfile.open(_NAME_PREDICTOR_FILE, 'r') as fp:
                conf = loads(fp.read().decode('utf-8'))
            with zfile.open(_NAME_GRAPH_FILE, 'r') as fp:
                graph = Graph()
                graph.parse(data=fp.read().decode('utf-8'))
            with zfile.open(_NAME_PIPELINE_FILE, 'r') as fp:
                pipeline = pkl.load(fp)
        pred = StwfsapyPredictor(
            graph=graph,
            concept_type_uri=_load_uri_ref(
                conf[_KEY_CONCEPT_TYPE_URI]),
            sub_thesaurus_type_uri=_load_uri_ref(
                conf[_KEY_THESAURUS_TYPE_URI]),
            thesaurus_relation_type_uri=_load_uri_ref(
                conf[_KEY_THESAURUS_RELATION_TYPE_URI]),
            thesaurus_relation_is_specialisation=(
                conf[_KEY_THESAURUS_RELATION_IS_SPECIALISATION]),
            remove_deprecated=conf[_KEY_REMOVE_DEPRECATED],
            langs=frozenset(conf[_KEY_LANGS]),
            handle_title_case=conf[_KEY_HANDLE_TITLE_CASE],
            extract_upper_case_from_braces=conf[
                _KEY_EXTRACT_UPPER_CASE_FROM_BRACES],
            extract_any_case_from_braces=conf[
                _KEY_EXTRACT_ANY_CASE_FROM_BRACES],
            expand_ampersand_with_spaces=conf[
                _KEY_EXPAND_AMPERSAND_WITH_SPACES],
            expand_abbreviation_with_punctuation=conf[
                _KEY_EXPAND_ABBREVIATION_WITH_PUNCTUATION],
            simple_english_plural_rules=conf[
                _KEY_SIMPLE_ENGLISH_PLURAL_RULES]
        )
        pred.dfa_ = dfa.Dfa.from_dict(conf[_KEY_DFA], str)
        pred.pipeline_ = pipeline
        pred.concept_map_ = conf[_KEY_CONCEPT_MAP]
        return pred


def _store_uri_ref(ref: URIRef) -> str:
    return ref.toPython()


def _load_uri_ref(uri: str) -> URIRef:
    return URIRef(uri)


def _handle_construction(
        con_state: construction.ConstructionState,
        concept: str,
        label: str):
    """Wrapper for construction that logs a warning
    in case of an exception.
    Uses concept and label as arguments for a more detailed message."""
    try:
        con_state.construct()
    except Exception:
        _logger.warning(
            f'Could not process label "{label}" of concept "{concept}".')
